#!/usr/bin/python
#
'''A test suite that implements an Atom parser using simplexmlparse.
Unfortunately, simplexmlparse doesn't actually implement enough features to
completely parse the complete Atom standard, but it can parse a significant
chunk of the Entry.

Atom standard: http://www.atompub.org/2005/08/17/draft-ietf-atompub-format-11.html

Test cases imported from: 
http://svn.sourceforge.net/viewcvs.cgi/feedvalidator/trunk/feedvalidator/testcases/atom/

Missing features:

- Handling xml:lang and xml:base. Currently, these attributes cause invalid
	identifier errors because Python identifiers cannot have colons (:).

- Handling arbitrary XML. SimpleParse will never be able to handle all of the
	XHTML standard, but there should be some way to match an arbitrary
	subtree of XML, either as the raw XML source, or as higher-level
	objects.

- Ignoring/parsing "extra" elements. Atom requires that unknown elements be
	silently ignored. It might be nice to collect these extra elements as
	arbitrary XML sequences.

http://www.atompub.org/2005/08/17/draft-ietf-atompub-format-11.html#rfc.section.6.3

- It would be nice to get SimpleParse raise errors if there is no text in
	elements that require them.

'''


import unittest

import simplexmlparse
from simpleparsetest import SimpleParseTestBase

import os

class TestFeed( SimpleParseTestBase ):
	def setUp(self):
		template = """
<feed xmlns="http://www.w3.org/2005/Atom" xmlns:parse="http://evanjones.ca/simplexmlparse">

	<!-- REQUIRED ELEMENTS -->
	<title parse:count="1" type="optional">title text</title>
	<id parse:count="1">URI</id>
	<updated parse:count="1">text</updated>

	<!-- RECOMMENDED ELEMENTS -->
	<author parse:count="*">
		<name parse:count="1">Full Name</name>
		<email>Email Address</email>
		<uri>Home Page of the Author</uri>
	</author>

	<link href="required http://example.org/" rel="optional"
		type="optional" hreflang="optional" title="optional" length="optional"
		parse:count="*"/>
	
	<!-- OPTIONAL ELEMENTS -->
	<category term="required" scheme="URI" label="human readable" parse:count="*"/>
	<contributor parse:count="*" parse:ref="author"/>
	<generator uri="optional" version="optional">text</generator>
	<icon>uri</icon>
	<logo>uri</logo>
	<rights>text</rights>
	<subtitle type="optional">text</subtitle>

	<entry parse:count="*">
		<!-- REQUIRED ELEMENTS -->
		<title parse:count="1" parse:ref="title"/>
		<id parse:count="1" parse:ref="title"/>
		<updated parse:count="1" parse:ref="updated"/>
	
		<!-- RECOMMENDED ELEMENTS -->
		<author parse:count="*" parse:ref="author"/>
		<content type="optional" src="URI">text</content>
		<link parse:count="*" parse:ref="link"/>
		<summary type="optional">text</summary>
		
		<!-- OPTIONAL ELEMENTS -->
		<category parse:count="*" parse:ref="category"/>
		<contributor parse:count="*" parse:ref="author"/>
		<published>text</published>
		<!-- TODO: <source></source> -->
		<rights parse:ref="rights"/>
	</entry>

</feed>"""

		self.parser = simplexmlparse.SimpleXMLParser( template )

	def testValid( self ):
		'''Tests that the feed can parse all the tests under the "atomtests/feed/valid" directory.'''
		for filename in os.listdir( 'atomtests/feed/valid' ):
			if not filename.endswith( '.atom' ):
				continue
			
			f = file( 'atomtests/feed/valid/' + filename )
			string = f.read()
			f.close()
			
			#~ print filename, 
			entry = self.parser.parse( string )
			#~ print 'PARSED'
	
	def testInvalid( self ):
		'''Tests that the feed fails to parse all the tests under the "atomtests/feed/invalid" directory.'''
		for filename in os.listdir( 'atomtests/feed/invalid' ):
			if not filename.endswith( '.atom' ):
				continue
			
			f = file( 'atomtests/feed/invalid/' + filename )
			string = f.read()
			f.close()
			
			#~ print filename, 
			try:
				entry = self.parser.parse( string )
			except simplexmlparse.ParseError, e:
				pass
				#~ print 'PARSE ERROR', e
			else:
				self.fail( "Expected a parse error!" )

class TestEntry( SimpleParseTestBase ):
	'''Tests a simplified Atom entry parser.'''
	def setUp(self):
		template = """
<entry xmlns="http://www.w3.org/2005/Atom" xmlns:simplexmlparse="http://evanjones.ca/simplexmlparse">

<!-- REQUIRED ELEMENTS -->
    <title simplexmlparse:count="1">title text</title>
    <id simplexmlparse:count="1">URI</id>
    <updated simplexmlparse:count="1">text</updated>

    <author simplexmlparse:count="+">
      <name simplexmlparse:count="1">Full Name</name>
      <email>Email Address</email>
      <uri>Home Page of the Author</uri>
    </author>
    
    <contributor simplexmlparse:count="*" simplexmlparse:ref="author"/>

    <content type="optional" src="URI">text</content>
    
    <link href="required http://example.org/"
        rel="optional" type="optional" hreflang="optional" title="optional" length="optional"
        simplexmlparse:count="*" />
    <summary type="optional">text</summary>
    <rights type="optional">text</rights>
    
    <category term="required" scheme="URI" label="human readable" simplexmlparse:count="*" />
    
    <published>text</published>
</entry>
"""
		self.parser = simplexmlparse.SimpleXMLParser( template )
	
	def testComplex( self ):
		doc = """
<entry xmlns="http://www.w3.org/2005/Atom">
    <title>Atom draft-07 snapshot</title>
    <link rel="alternate" type="text/html"
     href="http://example.org/2005/04/02/atom"/>
    <link rel="enclosure" type="audio/mpeg" length="1337"
     href="http://example.org/audio/ph34r_my_podcast.mp3"/>
    <id>tag:example.org,2003:3.2397</id>
    <updated>2005-07-11T12:29:29Z</updated>
    <published>2003-12-13T08:29:29-04:00</published>
    <author>
      <name>Mark Pilgrim</name>
      <uri>http://example.org/</uri>
      <email>f8dy@example.com</email>
    </author>
    <content type="text">[Update: The Atom draft is finished.]</content>
    <summary type="text">[Update: The Atom draft is finished.]</summary>
    <rights type="text">Some copyright statement</rights>
  </entry>"""
		
		entry = self.parser.parse( doc )
		
		assert( entry.title._text == 'Atom draft-07 snapshot' )
		assert( len( entry.link) == 2 )
		
		assert( entry.link[0].rel == 'alternate' )
		assert( entry.link[0].type == 'text/html' )
		assert( entry.link[0].href == 'http://example.org/2005/04/02/atom' )
		assert( entry.link[0].hreflang == None )
		assert( entry.link[0].length == None )
		
		assert( entry.link[1].rel == 'enclosure' )
		assert( entry.link[1].type == 'audio/mpeg' )
		assert( entry.link[1].href == 'http://example.org/audio/ph34r_my_podcast.mp3' )
		assert( entry.link[1].hreflang == None )
		assert( entry.link[1].length == "1337" )
		
		assert( entry.id._text == 'tag:example.org,2003:3.2397' )
		assert( entry.updated._text == '2005-07-11T12:29:29Z' )
		assert( entry.published._text == '2003-12-13T08:29:29-04:00' )
		
		assert( len( entry.author ) == 1 )
		assert( entry.author[0].name._text == 'Mark Pilgrim' )
		assert( entry.author[0].uri._text == 'http://example.org/' )
		assert( entry.author[0].email._text == 'f8dy@example.com' )
		
		assert( entry.content.type == 'text' )
		assert( entry.content._text == '[Update: The Atom draft is finished.]' )
		
		assert( entry.summary.type == 'text' )
		assert( entry.summary._text == '[Update: The Atom draft is finished.]' )
		
		assert( entry.rights.type == 'text' )
		assert( entry.rights._text == 'Some copyright statement' )
	
	def testValid( self ):
		'''Tests that the entry can parse all the tests under the "atomtests/entry/valid" directory.'''
		for filename in os.listdir( 'atomtests/entry/valid' ):
			if not filename.endswith( '.atom' ):
				continue
			
			f = file( 'atomtests/entry/valid/' + filename )
			string = f.read()
			f.close()
			
			#~ print filename, 
			entry = self.parser.parse( string )
			#~ print 'PARSED'
	
	def testInvalid( self ):
		'''Tests that the entry fails to parse all the tests under the "atomtests/entry/invalid" directory.'''
		for filename in os.listdir( 'atomtests/entry/invalid' ):
			if not filename.endswith( '.atom' ):
				continue
			
			f = file( 'atomtests/entry/invalid/' + filename )
			string = f.read()
			f.close()
			
			#~ print filename, 
			try:
				entry = self.parser.parse( string )
			except simplexmlparse.ParseError, e:
				pass
				#~ print 'PARSE ERROR'
			else:
				self.fail( "Expected a parse error!" )
	
	
		

unittest.main()
