#!/usr/bin/python

import xml.dom.minidom as minidom
import xml.sax
import unittest

import simplexmlparse

class SimpleParseTestBase( unittest.TestCase ):
	def tryParseError( self, constructor, constructorArgs, code ):
		
		try:
			if isinstance( constructorArgs, tuple ) or isinstance( constructorArgs, list ):
				obj = constructor( *constructorArgs )
			else:
				obj = constructor( constructorArgs )
		except simplexmlparse.ParseError, e:
			#~ print e
			assert( e.code == code )

			#~ p = e.domPath()
			#~ while path != None:
				#~ assert( p[-1] == path )
				#~ del p[-1]
				#~ path = path.parentNode
			#~ assert( len( p ) == 0 )
		else:
			self.fail( "Expected a simplexmlparse.ParseError" )

class TestParseError( unittest.TestCase ):
	def testLineColumnNumbers( self ):
		template = "<doc><a/><a/></doc>"
		try:
			parser = simplexmlparse.SimpleXMLParser( template )
		except simplexmlparse.ParseError, e:
			assert e.code == simplexmlparse.E_ELEMENT_CONFLICT
			assert e.line == 1
			assert e.column == 9
		
		template = "<doc>\n<a/>\n\t   <a/></doc>"
		try:
			parser = simplexmlparse.SimpleXMLParser( template )
		except simplexmlparse.ParseError, e:
			assert e.code == simplexmlparse.E_ELEMENT_CONFLICT
			assert e.line == 3
			assert e.column == 4

class TestDocument( SimpleParseTestBase ):
	def testEmpty( self ):
		# This template matches an empty document only
		template = "<doc>\n\t\r</doc>"	
		parser = simplexmlparse.SimpleXMLParser( template )
		
		docString = "<doc/>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		
		# comments are ignored
		docString = "<doc><!-- this is a comment --></doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		
		# Whitespace is ignored
		docString = "<doc>\n\r\t</doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		
		# Wrong doctag
		docString = "<document/>"
		self.tryParseError( parser.parse, docString, simplexmlparse.E_INVALID_DOCTAG )
		
		# text causes a parse error
		docString = "<doc>\n<!-- comment --> \t\rtext</doc>"
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_TEXT )
	
	def testText( self ):
		# This template matches an empty document or one with text
		template = "<doc>text</doc>"
		parser = simplexmlparse.SimpleXMLParser( template )
		
		# An empty document should work
		docString = "<doc/>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc._text == "" )
		
		# A document with text should work
		docString = "<doc>This element contains text</doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc._text == "This element contains text" )
		
		# A document with a child node should have an error
		docString = "<doc><child /></doc>"
		self.tryParseError( parser.parse, (docString,), simplexmlparse.E_NO_CHILDREN )
	
	def testChild( self ):
		# This template matches a document with one optional child.
		template = "<doc><child/></doc>"
		parser = simplexmlparse.SimpleXMLParser( template )
		
		docString = "<doc>\n\t<child/>\r\n</doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc.child != None )
		assert( doc.child.__class__.__name__ == 'child' )
		
		# An empty document should work
		docString = "<doc/>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc.child == None )
		
		# A document with text should not work
		docString = "<doc>\n<child>text</child>\n</doc>"
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_TEXT )
		
		# Multiple children should not work
		docString = "<doc>\n<child></child>\n<child/></doc>"
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_MULTIPLE )

	def testComplexChild( self ):
		template = "<doc><childOne><subChild /></childOne><childTwo /></doc>"
		parser = simplexmlparse.SimpleXMLParser( template )

		# An empty document should work
		docString = "<doc/>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc.childOne == None )
		assert( doc.childTwo == None )

		# Child one only
		docString = "<doc>\n\t<childOne/>\r\n</doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc.childOne != None )
		assert( doc.childTwo == None )
		assert( doc.childOne.__class__.__name__ == 'childOne' )
		assert( doc.childOne.subChild == None )
		
		# Child one and sub child
		docString = "<doc>\n\t<childOne><subChild /></childOne>\r\n</doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc.childOne != None )
		assert( doc.childTwo == None )
		assert( doc.childOne.__class__.__name__ == 'childOne' )
		assert( doc.childOne.subChild.__class__.__name__ == 'subChild' )
		
		# Child one, sub child, child Two
		docString = "<doc>\n<childTwo />\t<childOne><subChild /></childOne>\r\n</doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc.childOne != None )
		assert( doc.childTwo.__class__.__name__ == 'childTwo' )
		assert( doc.childOne.__class__.__name__ == 'childOne' )
		assert( doc.childOne.subChild.__class__.__name__ == 'subChild' )
		
		# Child one, child Two
		docString = "<doc>\n<childTwo />\t<childOne>\r\n</childOne>\r\n</doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc.childOne != None )
		assert( doc.childTwo.__class__.__name__ == 'childTwo' )
		assert( doc.childOne.__class__.__name__ == 'childOne' )
		assert( doc.childOne.subChild == None )
	
	def testIdentifiers( self ):
		# Leading underscores are not permitted
		template = "<doc><_child /></doc>"
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_INVALID_IDENTIFIER )
		
		# Not ASCII identifier
		template = u'<doc><\xe9lement /></doc>'.encode( 'utf-8' )
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_INVALID_IDENTIFIER )
		
		# Colons are not permitted
		template = "<doc><fake:namespace /></doc>"
		try:
			self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_INVALID_IDENTIFIER )
		# On some platforms this causes SAX to raise an "unbound namespace" exception
		# ignore it: that is fine as well
		except xml.sax.SAXParseException:
			pass
		
		template = "<doc><child attribute='optional' /></doc>"
		parser = simplexmlparse.SimpleXMLParser( template )
		
		# These strings are not permitted
		docString = u'<doc><\xe9lement /></doc>'.encode( 'utf-8' )
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_INVALID_CHILD )
		
		docString = '<doc><_child /></doc>'
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_INVALID_CHILD )
		
		docString = '<doc><child fake:attribute="value"/></doc>'
		try:
			self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_INVALID_ATTRIBUTE )
		# Ignore unbound attribute error
		except xml.sax.SAXParseException:
			pass
	
	def testReferences( self ):
		# Multiple elements with the same name!
		template = '<doc><a><entry>text</entry></a><b><entry>text</entry></b></doc>'
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_ELEMENT_CONFLICT )
		
		# Reference to non-existent element
		template = '''<doc xmlns:parse="http://evanjones.ca/simplexmlparse">
			<a><entry>text</entry></a><entry parse:ref="entry2"/></doc>'''
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_UNDEFINED_REF )
		
		# Reference on document
		template = '''<doc xmlns:parse="http://evanjones.ca/simplexmlparse" parse:ref="">
			<a><entry>text</entry></a></doc>'''
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_UNDEFINED_REF )
		
		# Reference with an attribute
		template = '''<doc xmlns:parse="http://evanjones.ca/simplexmlparse">
			<a><entry>text</entry></a><entry parse:ref="entry" attr1="value"/></doc>'''
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_REF_HAS_ATTRIBUTE )
		
		# Reference with content
		template = '''<doc xmlns:parse="http://evanjones.ca/simplexmlparse">
			<a><entry>text</entry></a><entry parse:ref="entry">content not permitted</entry></doc>'''
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_REF_NOT_EMPTY )
		
		template = '''<doc xmlns:parse="http://evanjones.ca/simplexmlparse">
			<a><entry>text</entry></a><entry parse:ref="entry"><content/></entry></doc>'''
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_REF_NOT_EMPTY )
		
		# Valid reference
		templateFixed = '''<doc xmlns:parse="http://evanjones.ca/simplexmlparse">
			<a><entry>text</entry></a><b><entry parse:ref="entry"/></b></doc>'''
		parser = simplexmlparse.SimpleXMLParser( templateFixed )
		
		document = '<doc><a><entry>a.entry</entry></a><b><entry>b.entry</entry></b></doc>'
		docObj = parser.parse( document )
		assert docObj.a.entry._text == 'a.entry'
		assert docObj.b.entry._text == 'b.entry'
		assert docObj.a.__class__ != docObj.b.__class__
		assert docObj.a.entry.__class__ == docObj.b.entry.__class__
		
		document = '<doc><a><entry>a.entry</entry></a><b/></doc>'
		docObj = parser.parse( document )
		assert docObj.a.entry._text == 'a.entry'
		assert docObj.b.entry == None
		
		# reference with count
		templateFixed = '''<doc xmlns:parse="http://evanjones.ca/simplexmlparse">
			<a><entry>text</entry></a><b><entry parse:ref="entry" parse:count="1"/></b></doc>'''
		
		parser = simplexmlparse.SimpleXMLParser( templateFixed )
		
		document = '<doc><a/><b><entry>b.entry</entry></b></doc>'
		docObj = parser.parse( document )
		assert docObj.a.entry == None
		assert docObj.b.entry._text == 'b.entry'
		
		document = '<doc><a><entry>a.entry</entry></a><b/></doc>'
		# entry is REQUIRED in b, not in a
		self.tryParseError( parser.parse, ( document ), simplexmlparse.E_MISSING )

class TestAttributes( SimpleParseTestBase ):
	def testSimpleAttributes( self ):
		template = "<doc><child attribute='blah blah' /></doc>"
		parser = simplexmlparse.SimpleXMLParser( template )
		
		# Empty still works
		docString = "<doc />"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc.child == None )
		
		# No attribute works (default is optional)
		docString = "<doc><child /></doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc.child.__class__.__name__ == "child" )
		assert( doc.child.attribute == None )
		
		# Specifying the attribute works
		docString = "<doc><child attribute='some string' /></doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc.child.__class__.__name__ == "child" )
		assert( doc.child.attribute == "some string" )
		
		# specifying a different attribute raises an error
		docString = "<doc><child otherAttribute='some string' /></doc>"
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_INVALID_ATTRIBUTE )
		
		# specifying a different attribute raises an error
		docString = "<doc invalid='error' />"
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_INVALID_ATTRIBUTE )
	
	
	def testRequiredAttributes( self ):
		template = "<doc required='required'><child required='optional' other='required' /></doc>"
		parser = simplexmlparse.SimpleXMLParser( template )
		
		# Empty fails: missing required attribute
		docString = "<doc />"
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_MISSING_ATTRIBUTE )
		
		# No child works
		docString = "<doc required='value' />"
		doc = parser.parse( docString )
		assert( doc.required == 'value' )
		assert( doc.child == None )
		
		# With child does not work (missing attribute)
		docString = "<doc required='value'><child required='actually optional' /></doc>"
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_MISSING_ATTRIBUTE )
		
		# With child works
		docString = "<doc required='value'><child other='value' /></doc>"
		doc = parser.parse( docString )
		assert( doc.required == 'value' )
		assert( doc.child.other == "value" )
		assert( doc.child.required == None )
		
		# Specifying all attributes works
		docString = "<doc required='value'><child required='actually optional' other='erg' /></doc>"
		doc = parser.parse( docString )
		assert( doc.required == 'value' )
		assert( doc.child.other == "erg" )
		assert( doc.child.required == 'actually optional' )
	
	def testAttributeTemplateErrors( self ):
		# There is a conflict between the attribute name and the child element
		template = "<doc><child attribute='blah'><attribute /></child></doc>"
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template ), simplexmlparse.E_CONFLICT )
		
		# But this is NOT a conflict: the attribute is on the child
		template = "<doc><child><attribute child='optional' /></child></doc>"
		parser = simplexmlparse.SimpleXMLParser( template )
		doc = parser.parse( '<doc><child><attribute child="hello world" /></child></doc>' )
		assert( doc.child.attribute.child == "hello world" )


class TestCounts( SimpleParseTestBase ):
	def testInvalidSpecialAttribute( self ):
		template = """<doc xmlns:xmlparse="http://evanjones.ca/simplexmlparse"><child xmlparse:invalid="true" /></doc>"""
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_INVALID_ATTRIBUTE )
		
	def testInvalidCounts( self ):
		template = """<doc xmlns:xmlparse="http://evanjones.ca/simplexmlparse" xmlparse:count="1" />"""
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_INVALID_ATTRIBUTE )
		
		template = """<doc xmlns:xmlparse="http://evanjones.ca/simplexmlparse">
			<child xmlparse:count="2" /> </doc>"""
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_COUNT )
		
		template = """<doc xmlns:simplexmlparse="http://evanjones.ca/simplexmlparse">
			<child simplexmlparse:count="some value" /> </doc>"""
		self.tryParseError( simplexmlparse.SimpleXMLParser, ( template, ), simplexmlparse.E_COUNT )

	def testRequiredChild( self ):
		template = """<doc xmlns:simplexmlparse="http://evanjones.ca/simplexmlparse">
				<child simplexmlparse:count="1" /></doc>"""
		parser = simplexmlparse.SimpleXMLParser( template )
		
		docString = "<doc>\n\t<child/>\r\n</doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( doc.child != None )
		assert( doc.child.__class__.__name__ == 'child' )
		
		# An empty document should not work
		docString = "<doc/>"
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_MISSING )
		
		# Multiple children should not work
		docString = "<doc>\n<child></child>\n<child/></doc>"
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_MULTIPLE )

	def testMultipleChildren( self ):
		template = """<doc xmlns:simplexmlparse="http://evanjones.ca/simplexmlparse">
				<child simplexmlparse:count="*" /></doc>"""
		parser = simplexmlparse.SimpleXMLParser( template )
		
		# Empty document works
		docString = "<doc />"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( len( doc.child ) == 0 )
		
		# One child should work
		docString = "<doc><child/></doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( len( doc.child ) == 1 )
		assert( doc.child[0].__class__.__name__ == 'child' )
		
		# Multiple children should work
		docString = "<doc>\n<child /> <child></child>\n<child/></doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( len( doc.child ) == 3 )
		assert( doc.child[0].__class__.__name__ == 'child' )
		assert( doc.child[1].__class__.__name__ == 'child' )
		assert( doc.child[2].__class__.__name__ == 'child' )
	
	def testOneOrMoreChildren( self ):
		template = """<doc xmlns:simplexmlparse="http://evanjones.ca/simplexmlparse">
				<child simplexmlparse:count="+" /></doc>"""
		parser = simplexmlparse.SimpleXMLParser( template )
		
		docString = "<doc>\n\t<child/>\r\n</doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( len( doc.child ) == 1 )
		assert( doc.child[0].__class__.__name__ == 'child' )
		
		# An empty document should not work
		docString = "<doc/>"
		self.tryParseError( parser.parse, ( docString, ), simplexmlparse.E_MISSING )
		
		# Multiple children should work
		docString = "<doc>\n<child /> <child></child>\n<child/></doc>"
		doc = parser.parse( docString )
		assert( doc.__class__.__name__ == "doc" )
		assert( len( doc.child ) == 3 )
		assert( doc.child[0].__class__.__name__ == 'child' )
		assert( doc.child[1].__class__.__name__ == 'child' )
		assert( doc.child[2].__class__.__name__ == 'child' )

if __name__ == "__main__":
	unittest.main()
